#include "RxTx.h"
#include <Arduino.h>


bool RxTx::_MODE_SENDING;
bool RxTx::_MODE_RECEIVING;
bool RxTx::_MODE_SNIFFING;
volatile bool RxTx::_MODE_AVAILABLE;

// RECEIVE
unsigned int RxTx::_INT_PIN_RX;

// SEND
unsigned int RxTx::_INT_PIN_TX;

// INTERUPTS
volatile unsigned int RxTx::_INT_RX_CNT;
volatile bool RxTx::_START_WITH_LOW_FLANK;

// TIMINGS
unsigned long RxTx::_TIME_MIN_OF_MIN;
unsigned long RxTx::_TIME_MIN_OF_MAX;
unsigned long RxTx::_TIME_MAX_OF_MIN;
unsigned long RxTx::_TIME_MAX_OF_MAX;
unsigned long RxTx::_TIME_MIN_OF_PERIOD;
unsigned long RxTx::_TIME_MAX_OF_PERIOD;
unsigned long RxTx::_PAKET_SIZE_TO_FILTER; // SNIFFER

// LOGIC
volatile bool RxTx::_FOUND_NEW_PAKET;
volatile unsigned int RxTx::_PAKET_LENGTH;
volatile unsigned int RxTx::_NUMBER_OF_PAKETS;
volatile unsigned int RxTx::_LOCAL_PAKET_COUNT;
volatile unsigned long RxTx::_TIME_LAST;
volatile unsigned long RxTx::_TIME_CURRENT;
volatile unsigned long RxTx::_PAKETFILTER_TIME_LAST;
volatile unsigned long RxTx::_PAKETFILTER_TIME_CURRENT;
volatile bool RxTx::_FIRST_HUGE_PAKET; // ONLY NEEDED IN COMBINATION WITH A RX UNIT

// DATA
unsigned int RxTx::_BUFFER_SIZE; // WE NEED ONE MORE TO CALCULATE THE DURATION OF THE LAST
volatile State* RxTx::_States;



RxTx::RxTx(unsigned int __INT_PIN_RX, unsigned int __INT_PIN_TX, unsigned int __BUFFER_SIZE) 
{
  _Init();

  _INT_PIN_RX = __INT_PIN_RX;
  _INT_PIN_TX = __INT_PIN_TX;
  _BUFFER_SIZE = __BUFFER_SIZE+1;
  _States = new State[_BUFFER_SIZE];

  Serial.begin(115200);
  delay(5000);
  Serial.println("Starting RxTx - signal analyser hack...");
}



void RxTx::_Init()
{
  _MODE_SENDING = false;
  _MODE_RECEIVING = false;
  _MODE_SNIFFING = false;
  _MODE_AVAILABLE = false;

  // RECEIVE
  _INT_PIN_RX = 10;// uC DUE
  //static const unsigned int _INT_PIN_RX = 2;// uC PRO MICRO - RX PIN (0) = INT 2

  // SEND
  _INT_PIN_TX = 11;

  // INTERUPTS
  _INT_RX_CNT = 0;
  _START_WITH_LOW_FLANK = false;

  // TIMINGS
  _TIME_MIN_OF_MIN = 0;
  _TIME_MIN_OF_MAX = 0;
  _TIME_MAX_OF_MIN = 0;
  _TIME_MAX_OF_MAX = 0;
  _TIME_MIN_OF_PERIOD = 0;
  _TIME_MAX_OF_PERIOD = 0;
  _PAKET_SIZE_TO_FILTER = 0; // SNIFFER

  // LOGIC
  _FOUND_NEW_PAKET = false;
  _PAKET_LENGTH = 0;
  _NUMBER_OF_PAKETS = 0;
  _LOCAL_PAKET_COUNT = 0;
  _TIME_LAST = 0;
  _TIME_CURRENT = 0;
  _PAKETFILTER_TIME_LAST = 0;
  _PAKETFILTER_TIME_CURRENT = 0;
  _FIRST_HUGE_PAKET = true; // ONLY NEEDED IN COMBINATION WITH A RX UNIT
}



unsigned int _DEBUG_CNT = 0;
void RxTx::Loop() 
{ 

  if(_MODE_SENDING)
  {
  	SendStatic();
  	SendStatic();
  	SendStatic();
  	SendStatic();
  	SendStatic();
  	SendStatic();
  	Serial.println("[INFO] Sent 5x ever 5 seconds");
  	delay(5000);
  }
  

  if(_MODE_RECEIVING)
  {
  	// PRINT RECEIVED VALUES
  	if(_MODE_AVAILABLE)
  	{
  	  // WAIT FOR FINISHED RECEIVE
  	  Serial.println("\nRECEIVING FINISHED!");
    
  	  // POSTDECODE DATA
  	  Serial.println("\nCALCULATING DURATIONS...");
  	  _CalculateDurations();
  	  
  	  // PRINT
  	  Serial.println("\nPRINTING...");
  	  _PrintStates(); 
  	  _PrintSLStates();
  	  
  	  // RESET ALL COUNTERS ETC.
  	  Serial.println("\nRESETTING...");
  	  _Reset();
    
  	  // READY
  	  Serial.println("\nREADY!");
  	}
  	
  	// SAVE PERFORMANCE
  	delay(1000);
    
  	//Serial.println(String(_INT_RX_CNT) + ",");
    
    if(_DEBUG_CNT==100)
    {
      Serial.print("\n.");
      _DEBUG_CNT = 0;
    }
    Serial.print(".");
    _DEBUG_CNT++;
  }
}



void RxTx::Sniff()
{
  _EnableINTrxSniff();
  _MODE_RECEIVING = true;
  _MODE_SNIFFING = true;
  _MODE_SENDING = false;
  _MODE_AVAILABLE = false;
}



void RxTx::SniffEstimatedPaketSize(unsigned int __PAKET_SIZE_TO_FILTER, unsigned long __TIME_MIN_OF_MIN, unsigned long __TIME_MAX_OF_MIN, unsigned long __TIME_MIN_OF_MAX, unsigned long __TIME_MAX_OF_MAX, unsigned long __TIME_MIN_OF_PERIOD, unsigned long __TIME_MAX_OF_PERIOD)
{
  _TIME_MIN_OF_PERIOD = __TIME_MIN_OF_PERIOD;
  _TIME_MAX_OF_PERIOD = __TIME_MAX_OF_PERIOD;

  _TIME_MIN_OF_MIN = __TIME_MIN_OF_MIN;
  _TIME_MAX_OF_MIN = __TIME_MAX_OF_MIN;

  _TIME_MIN_OF_MAX = __TIME_MIN_OF_MAX;
  _TIME_MAX_OF_MAX = __TIME_MAX_OF_MAX;

  _PAKET_SIZE_TO_FILTER = __PAKET_SIZE_TO_FILTER;

  _EnableINTrx();
  _MODE_RECEIVING = true;
  _MODE_SNIFFING = false;
  _MODE_SENDING = false;
  _MODE_AVAILABLE = false;
}



void RxTx::SendStatic()
{
  // ON
  // HEADER                                 // BODY
  //SL SS LL SL SL SL SL SL SL SL SL SL SS  LS LS LL SS LS LS LL SF
  //10 10 10 10 10 10 10 10 10 10 10 10 10  10 10 10 10 10 10 10 10
  _SendHigh(700);
  _SendLow(1400);
  
  _SendHigh(700);
  _SendLow(700);
  
  _SendHigh(1400);
  _SendLow(1400);
  
  _SendHigh(700);
  _SendLow(1400);
  
  _SendHigh(700);
  _SendLow(1400);
  
  _SendHigh(700);
  _SendLow(1400);
  
  _SendHigh(700);
  _SendLow(1400);
  
  _SendHigh(700);
  _SendLow(1400);
  
  _SendHigh(700);
  _SendLow(1400);
  
  _SendHigh(700);
  _SendLow(1400);
  
  _SendHigh(700);
  _SendLow(1400);
  
  _SendHigh(700);
  _SendLow(1400);
  
  _SendHigh(700);
  _SendLow(700);

  
  _SendHigh(1400);
  _SendLow(700);
  
  _SendHigh(1400);
  _SendLow(700);
  
  _SendHigh(1400);
  _SendLow(1400);

  _SendHigh(700);
  _SendLow(700);

  _SendHigh(1400);
  _SendLow(700);

  _SendHigh(1400);
  _SendLow(700);

  _SendHigh(1400);
  _SendLow(1400);

  _SendHigh(700);
  _SendLow(80100);
}



void RxTx::SendBuffer()
{
  Serial.println("\nWAITING...");
  _DisableINTrx(); // DISABLE RECEIVE INTERRUPTS
  delay(5000);
  
  Serial.println("\nSENDING...");
  for (unsigned int i = 0; i < _BUFFER_SIZE; i++)
  {
  	if(!(_States[i].Duration > _PAKET_SIZE_TO_FILTER) && (_States[i].Duration > 0))
  	{     
  	  Serial.println("[SEND] ID: " + String(_States[i].ID) + " Time: " + String(_States[i].Time) + " Flank: " + String(_States[i].Flank) + " Duration: " + String(_States[i].Duration));

  	  // SEND LOW
  	  if(_States[i].Flank == 0) 
  	  {
  		  _SendLow(_States[i].Duration);
  	  }
  	  
  	  // SEND HIGH
  	  if(_States[i].Flank == 1) 
  	  {
  		  _SendHigh(_States[i].Duration);
  	  }
  	}
  }

  // RESET ALL COUNTERS ETC.
  Serial.println("\nRESETTING...");
  _EnableINTrx();
  
  delay(5000);
  _Reset();
  
  // READY
  Serial.println("\nREADY!");
}



void RxTx::SendStates(State* __STATES, unsigned int __COUNT, unsigned int __REPEAT_COUNT)
{
  Serial.println("\nWAITING...");
  _DisableINTrx(); // DISABLE RECEIVE INTERRUPTS
  delay(5000);
  
  for(unsigned int j = 0; j < __REPEAT_COUNT; j++)
  {
    Serial.println("\nSENDING..." + String(j));
    for (unsigned int i = 0; i < __COUNT; i++)
    {
      if(!(__STATES[i].Duration > _PAKET_SIZE_TO_FILTER) && (__STATES[i].Duration > 0))
      {     
        Serial.println("[SEND] ID: " + String(__STATES[i].ID) + " Time: " + String(__STATES[i].Time) + " Flank: " + String(__STATES[i].Flank) + " Duration: " + String(__STATES[i].Duration));

        // SEND LOW
        if(__STATES[i].Flank == 0) 
        {
          _SendLow(__STATES[i].Duration);
        }
        
        // SEND HIGH
        if(__STATES[i].Flank == 1) 
        {
          _SendHigh(__STATES[i].Duration);
        }
      }
    }
  }

  // RESET ALL COUNTERS ETC.
  Serial.println("\nRESETTING...");
  _EnableINTrx();
  
  delay(5000);
  _Reset();
  
  // READY
  Serial.println("\nREADY!");
}



void RxTx::SendMessage(char* __MESSAGE_LS, unsigned long __TIME_DURATION_L, unsigned long __TIME_DURATION_S, unsigned long __TIME_DURATION_F, unsigned int __REPEAT_COUNT)
{
  bool _HIGH = false; // TOGGLE HIGH LOW

  for(unsigned int i = 0; i < __REPEAT_COUNT; i++)
  {
    unsigned int length = strlen(__MESSAGE_LS);
    Serial.print("\n[SENDING] Message with length " +  String(length) + ": ");
    for(unsigned int c = 0; c < length; c++)
    { 
      _HIGH = !_HIGH; // STRAT WITH HIGH FLANK
      if(__MESSAGE_LS[c] == 'S')
      {
        if(_HIGH) _SendHigh(__TIME_DURATION_S);
        else _SendLow(__TIME_DURATION_S);

        Serial.print(",S" + String((int)_HIGH));
      }

      if(__MESSAGE_LS[c] == 'L')
      {
        if(_HIGH) _SendHigh(__TIME_DURATION_L);
        else _SendLow(__TIME_DURATION_L);

        Serial.print(",L" + String((int)_HIGH));
      }

      if(__MESSAGE_LS[c] == 'F')
      {
        if(_HIGH) _SendHigh(__TIME_DURATION_F);
        else _SendLow(__TIME_DURATION_F);

        Serial.print(",F" + String((int)_HIGH));
      }
    }
  }
}




void RxTx::_CalculateDurations()
{
  for (unsigned int i = 1; i < _BUFFER_SIZE; i++)
  {
    _States[i-1].Duration = _States[i].Time - _States[i-1].Time;
  }
}



void RxTx::_PrintStates()
{
  if(!_MODE_SNIFFING) Serial.println("[PRINT] Number of Pakets: " + String(_NUMBER_OF_PAKETS) + " Paket Length: " + _PAKET_LENGTH);
  
  for (unsigned int i = 0; i < _BUFFER_SIZE; i++)
  {
  	if(_States[i].Time > 0)
  	{
  	  Serial.println("[PRINT] ID: " + String(_States[i].ID) + " Time: " + String(_States[i].Time) + " Flank: " + String(_States[i].Flank) + " Duration: " + String(_States[i].Duration));
  	}
  } 
}



void RxTx::_PrintSLStates()
{
  if(!_MODE_SNIFFING) 
  {
    Serial.println("\n[INFO] Number of Pakets: " + String(_NUMBER_OF_PAKETS) + " Paket Length: " + _PAKET_LENGTH);
    Serial.print("\n[INFO] Decoded Paket 1(H)ighs and 0(L)ows: ");
    bool HL = false;
    for (unsigned int i = 0; i < _PAKET_LENGTH; i++)
    {
      HL = !HL;
      Serial.print(String(HL));
    }
    for (unsigned int i = 0; i < _BUFFER_SIZE; i++)
    {
    	if(i % _PAKET_LENGTH == 0)
    	{
    	  Serial.print("\n[INFO] Decoded Paket S(hort)/L(ong)/F(inal) Period of toggeling Lows or Highs: ");
    	}
    	if(_States[i].Duration < _TIME_MAX_OF_MIN && _States[i].Duration > _TIME_MIN_OF_MIN)
    	{
    	  Serial.print("S");
    	  continue;
    	}
    	if(_States[i].Duration < _TIME_MAX_OF_MAX && _States[i].Duration > _TIME_MIN_OF_MAX)
    	{
    	  Serial.print("L");
    	  continue;
    	}
    	if(_States[i].Duration < _TIME_MAX_OF_PERIOD && _States[i].Duration > _TIME_MIN_OF_PERIOD)
    	{
    	  Serial.print("F");
    	  continue;
    	}
    }
  } 
}



void RxTx::_EnableINTrxSniff() 
{
  attachInterrupt(_INT_PIN_RX, _HandleInterruptChangeSniff, CHANGE);
}



void RxTx::_EnableINTrx() 
{
  attachInterrupt(_INT_PIN_RX, _HandleInterruptChange, CHANGE);
}



void RxTx::_DisableINTrx() 
{
  detachInterrupt(_INT_PIN_RX);
}



void RxTx::_HandleInterruptChangeSniff() 
{  
  if(_MODE_AVAILABLE) return;// NOT INTERESTED IN INTERRUPTS WHILE WE ANALYSE THEM
  
  // FETCH CURRENT PAKET NUMBER
  unsigned int _GLOBAL_CNT = _INT_RX_CNT;
  if(_GLOBAL_CNT == _BUFFER_SIZE)
  {
  	_INT_RX_CNT = 0;
  	_MODE_AVAILABLE = true;
  	return;
  }
  
  // UPDATE PAKET NUMBERS
  _INT_RX_CNT++; // GLOABL
  
  // TEST FOR UPDATE AVAILIBILITY
  _TIME_CURRENT = micros();
  
  // SAVE GOOD PAKET VALUE
  _START_WITH_LOW_FLANK = !_START_WITH_LOW_FLANK; 
  _States[_GLOBAL_CNT].ID = _GLOBAL_CNT;
  _States[_GLOBAL_CNT].Time = _TIME_CURRENT;
  _States[_GLOBAL_CNT].Flank = (int)_START_WITH_LOW_FLANK;
}



void RxTx::_HandleInterruptChange() 
{   
  if(_MODE_AVAILABLE) return;// NOT INTERESTED IN INTERRUPTS WHILE WE ANALYSE THEM


// FILTER NEW PAKETS
  unsigned long _CURRENT_DURATION = _GetCurrentDuration(); // CHECK IF WE FOUND A VALID PAKET START POINT
  if((_CURRENT_DURATION > _PAKET_SIZE_TO_FILTER-1000) && (_CURRENT_DURATION < _PAKET_SIZE_TO_FILTER + 1000))
  {
  	// ?? FETCH THE FIRST AS IT IS A CALC MISTAKE ??
  	//if(_FIRST_HUGE_PAKET){_FIRST_HUGE_PAKET = false; return;}
  	
  	if(_FOUND_NEW_PAKET == true) // WE ALREADY FOUND A NEW VALID PAKET
  	{  
  	  // NOW WE KNOW THE PAKET LENGTH
  	  if(_PAKET_LENGTH == 0)
  	  {
  		  _PAKET_LENGTH = _LOCAL_PAKET_COUNT;
  	  }
  	}
  	_LOCAL_PAKET_COUNT = 0;
  	_NUMBER_OF_PAKETS++;
  	_FOUND_NEW_PAKET = true;
  }

  
// PROCESS
  if(_FOUND_NEW_PAKET)
  {
  	// FETCH CURRENT PAKET NUMBER
  	unsigned int _GLOBAL_CNT = _INT_RX_CNT;// FREE INTERRUPT ACCESS ASAP/HERE
  	_INT_RX_CNT++; 
  	
  	// EXIT IF BUFFER FULL || NUMBER OF PAKETS EXEEDS BUFFER || COUNTER EXEEDS PAKET LENGTH
    if((_GLOBAL_CNT == _BUFFER_SIZE)||(_PAKET_LENGTH * _NUMBER_OF_PAKETS > _BUFFER_SIZE)||(_PAKET_LENGTH > 0 && _LOCAL_PAKET_COUNT > _PAKET_LENGTH))
  	{
  		// SAVE TEH LAST GOOD PAKET VALUE
  		_START_WITH_LOW_FLANK = !_START_WITH_LOW_FLANK; 
  		_States[_GLOBAL_CNT].ID = _GLOBAL_CNT;
  		_States[_GLOBAL_CNT].Time = micros();
  		_States[_GLOBAL_CNT].Flank = (int)_START_WITH_LOW_FLANK;
  	
  	  // FINISHED RECEIVING - DO NOT ACCEPT ADDITIONAL INTERRUPTS
  	  _MODE_AVAILABLE = true;
  	  _FOUND_NEW_PAKET = false;
  	  return;
  	}

  	_LOCAL_PAKET_COUNT++; // LOCAL - ALWAYS COUNTING UP TO PAKET LENGTH TO GET THE LAST PAKET
  	
  	// SAVE GOOD PAKET VALUE
  	_START_WITH_LOW_FLANK = !_START_WITH_LOW_FLANK; 
  	_States[_GLOBAL_CNT].ID = _GLOBAL_CNT;
  	_States[_GLOBAL_CNT].Time = micros();
  	_States[_GLOBAL_CNT].Flank = (int)_START_WITH_LOW_FLANK;
  }
}



void RxTx::_Reset()
{ 
  // RESET LOGIC
  _LOCAL_PAKET_COUNT = 0;
  _NUMBER_OF_PAKETS = 0;
  _PAKET_LENGTH = 0;
  _MODE_AVAILABLE = false;
  _INT_RX_CNT = 0;
  _TIME_CURRENT = 0;
  _TIME_LAST = 0;
  
  // INIT BUFFER
  for (unsigned int i = 0; i < _BUFFER_SIZE; i++)
  {
  	_States[i].ID = 0;
  	_States[i].Time = 0;
  	_States[i].Duration = 0;
  	_States[i].Flank = 2;
  }
}



unsigned long RxTx::_GetCurrentDuration()
{
  _PAKETFILTER_TIME_CURRENT = micros();
  unsigned long _DURATION = 0;
  if(_PAKETFILTER_TIME_LAST > 0) _DURATION = _PAKETFILTER_TIME_CURRENT - _PAKETFILTER_TIME_LAST;
  _PAKETFILTER_TIME_LAST = _PAKETFILTER_TIME_CURRENT;
  return _DURATION;
}



void RxTx::_SendLow(unsigned long _time)
{
  digitalWrite(_INT_PIN_TX, LOW); 
  delayMicroseconds(_time);
}



void RxTx::_SendHigh(unsigned long _time)
{
  digitalWrite(_INT_PIN_TX, HIGH); 
  delayMicroseconds(_time);  
}
