#ifndef RXTX_H
#define RXTX_H



typedef struct 
{
 unsigned int ID;
 unsigned int Flank;
 unsigned long Time;
 unsigned long Duration;
} State;



// THIS LIB WAS INTEGRATED WITH THE PURPOSE TO WORK WITH HOLTEK 4xR01T3v130
// http://www.holtek.com/pdf/consumer/4xR01T3v130.pdf
class RxTx
{
private:
  static bool _MODE_SENDING;
  static bool _MODE_RECEIVING;
  static bool _MODE_SNIFFING;
  static volatile bool _MODE_AVAILABLE;

  /*
  // ATTENTION
  // uC DUE - RX Pin any digital PIN, TX Pin any digital PIN
  // uC SparkfunPro Micro 3V3 - RX Pin 2, TX Pin any digital PIN
	// See: https://www.arduino.cc/en/Reference/AttachInterrupt
	Board	Digital Pins Usable For Interrupts
	Uno, Nano, Mini, other 328-based:	2, 3
	Mega, Mega2560, MegaADK:	2, 3, 18, 19, 20, 21
	Micro, Leonardo, other 32u4-based:	0, 1, 2, 3, 7
	Zero:	all digital pins, except 4
	Due:	all digital pins
  */
  // RECEIVE
  static unsigned int _INT_PIN_RX;

  // SEND
  static unsigned int _INT_PIN_TX;

  // INTERUPTS
  static volatile unsigned int _INT_RX_CNT;
  static volatile bool _START_WITH_LOW_FLANK;

  // TIMINGS
  static unsigned long _TIME_MIN_OF_MIN;
  static unsigned long _TIME_MIN_OF_MAX;
  static unsigned long _TIME_MAX_OF_MIN;
  static unsigned long _TIME_MAX_OF_MAX;
  static unsigned long _TIME_MIN_OF_PERIOD;
  static unsigned long _TIME_MAX_OF_PERIOD;
  static unsigned long _PAKET_SIZE_TO_FILTER; // SNIFFER

   // LOGIC
  static volatile bool _FOUND_NEW_PAKET;
  static volatile unsigned int _PAKET_LENGTH;
  static volatile unsigned int _NUMBER_OF_PAKETS;
  static volatile unsigned int _LOCAL_PAKET_COUNT;
  static volatile unsigned long _TIME_LAST;
  static volatile unsigned long _TIME_CURRENT;
  static volatile unsigned long _PAKETFILTER_TIME_LAST;
  static volatile unsigned long _PAKETFILTER_TIME_CURRENT;
  static volatile bool _FIRST_HUGE_PAKET; // ONLY NEEDED IN COMBINATION WITH A RX UNIT
  
  // DATA
  static unsigned int _BUFFER_SIZE; // WE NEED ONE MORE TO CALCULATE THE DURATION OF THE LAST
	static volatile State* _States;



public:
	// Example for Arduino DUE
	// INIT RXTX SIGNAL ANALYSER PIN RX , PIN TX, BUFFER SIZE
  RxTx(unsigned int __INT_PIN_RX=10, unsigned int __INT_PIN_TX=11, unsigned int __BUFFER_SIZE=1024);
  
  // SNIFF FOR INCOMMING PAKETS TO FIND OUT THE PAKETS SYNC PERIOD/BREAK AND TIMINGS
  // e.g. rxtx->Sniff();
  static void Sniff();

  // Example for HT48R01T3
  // SNIFF WITH PAKETS PERIOD SIZE AND TIMINGS
  // FILTER PAKET STRUCTURE WITH SYNC PERIOD
  // e.g. rxtx->SniffEstimatedPaketSize(83000);

  // FILTER PAKET STRUCTURE WITH COMPLETE REVERSED INFORMATION
  // SYNC_AVG, MIN_OF_MIN, MAX_OF_MIN, MIN_OF_MAX, MAX_OF_MAX, MIN_OF_SYNC, MAX_OF_SYNC
  // e.g. rxtx->SniffEstimatedPaketSize(83000, 600, 800, 1300, 1500, 75000, 85000);
  static void SniffEstimatedPaketSize(unsigned int __PAKET_SIZE_TO_FILTER=83000, unsigned long _TIME_MIN_OF_MIN=600, unsigned long _TIME_MAX_OF_MIN=800, unsigned long _TIME_MIN_OF_MAX=1300, unsigned long _TIME_MAX_OF_MAX=1500, unsigned long _TIME_MIN_OF_PERIOD=75000, unsigned long _TIME_MAX_OF_PERIOD=85000);
  
  // TODO
  static State* GetSniffedPakets(unsigned int __PAKET_SIZE_TO_FILTER=83000, unsigned long _TIME_MIN_OF_MIN=600, unsigned long _TIME_MAX_OF_MIN=800, unsigned long _TIME_MIN_OF_MAX=1300, unsigned long _TIME_MAX_OF_MAX=1500, unsigned long _TIME_MIN_OF_PERIOD=75000, unsigned long _TIME_MAX_OF_PERIOD=85000);
  
  // TODO
  // ONLY IF YOU HAVE A RECEIVED BUFFER AND DID NOT DELETE IT - SO BE AWARE TO GENERATE A CORRECT "static volatile State* _States;" BUFFER
  static void SendBuffer();

	// TODO
  // GENERATE A CORRECT "State*" BUFFER
  static void SendStates(State* __STATES, unsigned int __COUNT, unsigned int __REPEAT_COUNT);

  // WORKS
  // e.g. message "SLSSLLSLSLSLSLSLSLSLSLSLSSLSLSLLSSLSLSLLSF" , WITH S = 700, L = 1400, F = 81000, repeat sending for __REPEAT_COUNT times
  static void SendMessage(char* __MESSAGE_LS, unsigned long __TIME_DURATION_L=1400, unsigned long __TIME_DURATION_S=700, unsigned long __TIME_DURATION_F=81000, unsigned int __REPEAT_COUNT=4);
  // WORKS
  static void SendStatic();

  // MUST BE USED TO RECEIVE SOMETHING
  static void Loop();



private:
  static void _Init();
  static void _CalculateDurations();
  static void _PrintStates();
  static void _PrintSLStates();
  static void _EnableINTrxSniff();
  static void _EnableINTrx();
  static void _DisableINTrx();
  static void _HandleInterruptChangeSniff();
  static void _HandleInterruptChange();
  static void _AddPaket(unsigned long _TIME_CURRENT, unsigned long _DURATION);
  static void _Reset();
  static unsigned long _GetCurrentDuration();
  static void _SendLow(unsigned long _time);
  static void _SendHigh(unsigned long _time);
};

#endif //RXTX_H
